# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import logging
import time
import sys

from psycopg2.extensions import TransactionRollbackError

import storm
from storm.locals import create_database, Desc, Store

from offspring.daemon import LexbuilderDaemon
from offspring.master.exceptions import SlaveCommunicationError
from offspring.master.models import BuildRequest, Lexbuilder
from offspring.master import notifications
from offspring.slave import LexbuilderSlave
from offspring.master.notifications import mail_notification
from offspring.config import setup_logging


class LexbuilderMaster(LexbuilderDaemon):

    def __init__(self, config, db_store=None, mailer=mail_notification):
        self.mailer = mailer
        self.db_store = db_store
        self.is_online = False
        self.config = config
        self.setup_logging()
        self.setup_database()
        LexbuilderDaemon.__init__(self)

    def setup_logging(self):
        if self.config.has_option("master", "logfile"):
            setup_logging(path=self.config.get("master", "logfile"),
                          level=logging.INFO,
                          fmt="%(asctime)s %(levelname)-8s %(message)s",
                          datefmt="%a, %d %b %Y %H:%M:%S")
        logging.info("Starting up offspring master")

    def setup_database(self):
        if not self.db_store:
            try:
                db_uri = self.config.get("master", "db")
                self.database = create_database(db_uri)
                self.db_store = Store(self.database)
            except storm.exceptions.OperationalError, e:
                logging.exception("Unable to connect to database %s." % db_uri)
                if not self.is_online:
                    raise e
            except storm.exceptions.DisconnectionError, e:
                logging.error("Disconnected from database %s." % db_uri)
                self.db_store = None
            if self.db_store:
                logging.info("Connected to database")
        self.is_online = True

    def run(self):
        poll_frequency = self.config.getint("master", "poll_frequency")
        while self.is_online:
            if not self.db_store:
                self.setup_database()
            else:
                try:
                    self.db_store.flush()
                    self.scanSlaves()
                    self.dispatchBuilds()
                except storm.exceptions.DisconnectionError:
                    logging.exception("Master was disconnected from database.")
                    if self.db_store:
                        self.db_store.rollback()
                        self.db_store.close()
                    self.db_store = None
                    logging.info(
                        "Waiting {0} seconds and attempting reconnect.".format(
                          poll_frequency))
                except TransactionRollbackError:
                    logging.exception("TransactionRollbackError occurred... ")
                    logging.info("Performing rollback to recover from error.")
                    self.db_store.rollback()
                except Exception:
                    logging.exception(
                        "Uncaught error propagated to main execution loop.")
                    notifications.notify_crash(
                        "Lexbuilder Master (Build Queue Dispatcher)",
                         self.config,
                         mailer=self.mailer)
                    logging.error(
                        "SHUTTING DOWN: Fatal error has occurred.")
                    sys.exit(1)
            time.sleep(poll_frequency)
        logging.info("No longer online, closing database connections")
        self.db_store.commit()
        self.db_store.close()

    def stop(self):
        self.is_online = False

    def _get_builders(self):
        return self.db_store.find(Lexbuilder, Lexbuilder.is_retired == False)

    def scanSlaves(self):
        """
        Scan the currently known build slaves.
        """
        for builder in self._get_builders():
            try:
                old_state, new_state = builder.scan()
            except SlaveCommunicationError, issue:
                logging.error(
                    "Scan of %s failed. %s" % (builder.name, issue))
                builder.is_okay = False
                self.db_store.commit()
                continue
            self.db_store.commit()
            logging.debug("%s: %s->%s" % (builder.name, old_state, new_state))

            if (builder.current_state == LexbuilderSlave.STATE_IDLE and
                builder.current_job != None):
                try:
                    if builder.getBuildResult():
                        logging.info("%s has finished building on %s" %
                            (builder.current_job.project.name, builder.name))
                        #XXX: Notification stuff needs work.
                        try:
                            notifications.notify_buildresult(
                                builder.current_job, self.config,
                                mailer=self.mailer)
                        except:
                            logging.exception(
                                "Attempt to send build result notification "
                                "e-mail failed.")
                            logging.error(
                                "PROBLEM DETECTED: A non-fatal problem has "
                                "been detected. See above for details.")
                            try:
                                notifications.notify_problem(
                                  "Lexbuilder Master (Build Queue Dispatcher)",
                                  "unable to send build result notification",
                                  self.config,
                                  mailer=self.mailer)
                            except:
                                logging.exception(
                                    "Attempt to send problem notification "
                                    "e-mail failed.")
                                logging.error(
                                    "PROBLEM DETECTED: A non-fatal problem has"
                                    " been detected. See above for details.")
                        builder.current_job = None
                        self.db_store.commit()
                    else:
                        logging.error(
                            "Failed to gather build result from %s" %
                            (builder.name))
                        logging.info(
                            "Deleting %s build record %s for lost build on "
                            "%s" % (builder.current_job.project.name,
                                    builder.current_job, builder.name))
                        self.db_store.remove(builder.current_job)
                        builder.current_job = None
                        self.db_store.commit()
                except SlaveCommunicationError, issue:
                    logging.error(
                        "Failed to gather build results from %s. %s" %
                        (builder.name, issue))
                    builder.is_okay = False
                    self.db_store.commit()
                    continue
            self.db_store.flush(),
        logging.debug("Scan cycle completed")

    def dispatchBuilds(self):
        #XXX: ARCH_MACHINE_SUPPORT is a bit of hack
        ARCH_MACHINE_SUPPORT = {
                u"i386": [u"i386", u"i686", u"x86_64"],
                u"lpia": [u"i386", u"i686", u"x86_64"],
                u"amd64": [u"x86_64"],
                u"armel": [u"armv7l"],
                u"armhf": [u"armv7l"],
        }
        build_requests = self.db_store.find(BuildRequest).order_by(
            Desc(BuildRequest.score))
        for request in build_requests:
            project = request.project
            logging.debug(
                "Project %s needs building (priority: %s)." %
                (project.name, project.priority))
            builder_types = ARCH_MACHINE_SUPPORT[unicode(project.arch)]
            builder = self.db_store.find(Lexbuilder,
                    Lexbuilder.current_state == LexbuilderSlave.STATE_IDLE,
                    Lexbuilder.is_active == True,
                    Lexbuilder.is_okay == True,
                    Lexbuilder.is_retired == False,
                    Lexbuilder.machine_type.is_in(builder_types)
                ).order_by(Lexbuilder.id).first()
            if builder is not None:
                try:
                    job = builder.build(request)
                    if job != None:
                        self.db_store.add(job)
                        self.db_store.remove(request)
                        self.db_store.commit()
                        logging.info(
                            "%s has been dispatched to %s" %
                            (project.name, builder.name))
                    else:
                        logging.error(
                            "Error when asking %s to build %s" %
                            (builder.name, project.name))
                except SlaveCommunicationError, issue:
                    logging.error(
                        "Failed to dispatch %s build to %s. %s" %
                        (project.name, builder.name, issue))
                    builder.is_okay = False
                    self.db_store.commit()
                    continue
            else:
                logging.debug(
                    "Unable to find builder to build %s" % project.name)
        self.db_store.flush()
