# Copyright 2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime

from storm import tz

from offspring.master.utils import (
    build_active_projects, process_pending_build_orders)
from offspring.master.models import (
    BuildRequest,
    Project,
    DailyBuildOrder,
    DailyBuildOrderProject,
    LaunchpadProject,
    LaunchpadProjectMilestone)
from offspring.master.scripts import sync_launchpad_project_milestones

from offspring.master.tests.helpers import (
    BaseOffspringMasterTestCase,
    FakeLaunchpadProject,
    FakeLaunchpadMilestone
)

class TestSyncLaunchpadProjectMilestones(BaseOffspringMasterTestCase):

    def test_sync_project_milestones(self):
        """
        Test milestone syncing succeeds.
        """
        lp_project = LaunchpadProject()
        lp_project.name = u"test-project"
        self.db_store.add(lp_project)

        lp_project_milestone = LaunchpadProjectMilestone(
            u"test-project-1.0", u"test-project-1.0", lp_project)
        self.db_store.add(lp_project_milestone)

        launchpad = self.mocker.mock()
        launchpad.projects["test-project"]
        fake_lp_milestone = FakeLaunchpadMilestone(u"test-project-1.0")
        self.mocker.result(FakeLaunchpadProject(u"test-project", fake_lp_milestone))
        self.mocker.replay()

        self.assertIs(
            sync_launchpad_project_milestones.sync_milestones(
                launchpad, self.db_store), None)


    def test_sync_project_milestones_handles_bad_milestone(self):
        """
        Test milestone syncing does not abort with bad milestone objects.
        """
        lp_project = LaunchpadProject()
        lp_project.name = u"test-project"
        self.db_store.add(lp_project)

        lp_project_milestone = LaunchpadProjectMilestone(
            u"test-project-1.0", u"test-project-1.0", lp_project)
        self.db_store.add(lp_project_milestone)

        launchpad = self.mocker.mock()
        launchpad.projects[u"test-project"]
        fake_lp_milestone = FakeLaunchpadMilestone(u"test-project-1.0")
        delattr(fake_lp_milestone, "name")
        fake_lp_project = FakeLaunchpadProject(
            u"test-project", fake_lp_milestone)
        self.mocker.result(fake_lp_project)
        self.mocker.replay()

        self.assertIs(
            sync_launchpad_project_milestones.sync_milestones(
                launchpad, self.db_store), None)


class TestBuildActiveProjects(BaseOffspringMasterTestCase):

    def test_build_active_projects(self):
        """
        build_active_projects should initiate a BuildRequest for every active
        Project.
        """
        self.assertEqual(0, self.db_store.find(BuildRequest).count())
        build_active_projects(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Nightly build", build_request.reason)

    def test_build_active_projects_ignores_inactive_projects(self):
        """
        build_active_projects should ignore any projects that are not marked as
        active.
        """
        project = Project(u"inactive-project")
        project.config_url = u"http://example.com/"
        project.series = u"inactive-series"
        project.suite = u"inactive-suite"
        project.arch = u"i386"
        project.is_active = False
        self.db_store.add(project)

        self.assertEqual(0, self.db_store.find(BuildRequest).count())
        build_active_projects(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Nightly build", build_request.reason)


class TestProcessPendingBuildOrders(BaseOffspringMasterTestCase):

    def test_process_pending_build_orders(self):
        """
        process_pending_build_orders should identify DailyBuildOrders that are
        scheduled at the same hour, and initiate BuildRequests for them.
        """
        current_hour = datetime.now(tz.tzlocal()).hour
        build_order = DailyBuildOrder(self.project.name, current_hour)
        self.db_store.add(build_order)
        build_order_project = DailyBuildOrderProject()
        build_order_project.dailybuildorder_id = build_order.id
        build_order_project.project_id = self.project.name
        self.db_store.add(build_order_project)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())

        process_pending_build_orders(self.db_store)
        build_request = self.db_store.find(BuildRequest).one()
        self.assertEqual(self.project, build_request.project)
        self.assertEqual(u"Daily build", build_request.reason)

    def test_process_pending_build_orders_ignores_orders_different_times(self):
        """
        process_pending_build_orders should ignore DailyBuildOrders that are
        not scheduled at the current time.
        """
        build_hour = datetime.now(tz.tzlocal()).hour + 1
        build_order = DailyBuildOrder(self.project.name, build_hour)
        self.db_store.add(build_order)
        build_order_project = DailyBuildOrderProject()
        build_order_project.dailybuildorder_id = build_order.id
        build_order_project.project_id = self.project.name
        self.db_store.add(build_order_project)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())

        process_pending_build_orders(self.db_store)
        self.assertEqual(0, self.db_store.find(BuildRequest).count())
