# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from datetime import datetime, timedelta, date

from django.test import TestCase
from mocker import MockerTestCase

from django_group_access.models import AccessGroup

from offspring.enums import ProjectBuildStates

from offspring.web.queuemanager.models import (
    ProjectGroup, get_completed_builds, get_updated_builders, Lexbuilder)
from offspring.web.queuemanager.tests.factory import factory
from offspring.web.queuemanager.tests.helpers import (
    create_build_results_with_durations
)

class ProjectTests(TestCase):

    def test_average_build_time(self):
        """
        Project.average_build_time should use the BuildResult metric.
        """
        project = factory.make_project()
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS)
        self.assertEqual(50 * 60.0, project.average_build_time)


class LexbuilderTests(TestCase, MockerTestCase):

    def test_estimated_completion(self):
        """
        Lexbuilder.estimated_completion_time should be calculated as:
            average build time - (current time - start time)
        """
        project = factory.make_project()

        create_build_results_with_durations(
            project=project, durations=[25, 25],
            result=ProjectBuildStates.SUCCESS)

        started_at = datetime.now() - timedelta(minutes=10)
        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING,
            started_at=started_at)
        builder = factory.make_lexbuilder(current_job=build_result)

        datetime_mock = self.mocker.replace(datetime)
        datetime_mock.now()
        self.mocker.result(started_at + timedelta(minutes=5))
        self.mocker.replay()

        # Average time = 25 minutes
        # Started 5 minutes ago (started_at + timedelta(minutes=5))
        # 20 minutes remaining

        self.assertEqual(20 * 60, builder.estimated_completion)

    def test_estimated_completion_with_no_average_time(self):
        """
        If the average cannot be determined because there are no builds, we
        should get "Unknown" as the estimated completion time.
        """
        project = factory.make_project()

        started_at = datetime.now() - timedelta(minutes=10)
        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING,
            started_at=started_at)

        builder = factory.make_lexbuilder(current_job=build_result)

        self.assertEqual("Unknown", builder.estimated_completion)

    def test_estimated_completion_with_no_current_build(self):
        """
        If there is no current build, then we should get "No build in
        progress" for the estimated completion time.
        """
        builder = factory.make_lexbuilder(current_job=None)
        self.assertEqual("No build in progress", builder.estimated_completion)

    def test_estimated_completion_negative_time(self):
        """
        If the current elapsed time is higher than the average, we should get
        "ASAP" as the estimated completion.
        """
        project = factory.make_project()

        create_build_results_with_durations(
            project=project, durations=[5, 5],
            result=ProjectBuildStates.SUCCESS)

        build_result = factory.make_build_result(
            project=project, result=ProjectBuildStates.PENDING)
        started_at = datetime.now() - timedelta(minutes=10)
        build_result.started_at = started_at
        build_result.save()

        builder = factory.make_lexbuilder(current_job=build_result)

        datetime_mock = self.mocker.replace(datetime)
        datetime_mock.now()
        self.mocker.result(started_at + timedelta(minutes=10))
        self.mocker.replay()

        # Average time = 5 minutes
        # Started 10 minutes ago (started_at + timedelta(minutes=10))
        self.assertEqual("ASAP", builder.estimated_completion)

    def test_average_load(self):
        """
        Lexbuilder.average_load should calculate the total time used by builds
        as a percentage of the time available.
        """
        builder = factory.make_lexbuilder()
        project = factory.make_project()
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS,
            build_date=datetime.now() - timedelta(days=2),
            builder=builder)
        # 50 minutes * 60 seconds + 50 minutes * 60 seconds / (24 * 60 * 60)
        # = 0.0694 = 6.94%
        self.assertEqual(6.94, round(builder.average_load, 2))

    def test_average_load_buildresults_today(self):
        """
        Lexbuilder.average_load should ignore buildresults from today.
        """
        builder = factory.make_lexbuilder()
        project = factory.make_project()

        create_build_results_with_durations(
            project, [50, 50], build_date=datetime.now().date(),
            result=ProjectBuildStates.SUCCESS, builder=builder)
        self.assertEqual(None, builder.average_load)


class ProjectGroupTests(TestCase):

    def test_average_build_time(self):
        """
        ProjectGroup.average_build_time should use the BuildResult metric.
        """
        project_group = factory.make_project_group()
        project = factory.make_project(project_group=project_group)
        create_build_results_with_durations(
            project, [10, 20], result=ProjectBuildStates.SUCCESS)
        self.assertEqual(15 * 60.0, project_group.average_build_time)

    def test_accessible_by_user(self):
        """
        ProjectGroup.accessible_by_user should return a list of
        ProjectGroups accessible to the user, these are ProjectGroups with a
        Project that is accessible to the user.
        """
        user = factory.make_user()

        group1 = AccessGroup.objects.create(name='Group1')
        group2 = AccessGroup.objects.create(name='Group2')
        group1.members.add(user)

        project_group1 = factory.make_project_group()
        project_group2 = factory.make_project_group()

        project1 = factory.make_project(project_group=project_group1)
        project2 = factory.make_project(project_group=project_group1)
        project3 = factory.make_project(project_group=project_group2)

        project1.access_groups.add(group1)
        project2.access_groups.add(group2)
        project3.access_groups.add(group2)

        self.assertEqual([project_group1],
                         list(ProjectGroup.objects.accessible_by_user(user)))



class UtilityFunctionTests(MockerTestCase, TestCase):

    def test_get_completed_builds(self):
        """
        get_completed_builds(timedelta(hours=1)) should return all BuildResults
        which have finished within the last hour.
        """
        project = factory.make_project()
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS)
        started_at = datetime.utcnow() - timedelta(hours=2)
        factory.make_build_result(
            project=project, requestor=None,
            result=ProjectBuildStates.SUCCESS,
            finished_at=started_at + timedelta(minutes=30),
            builder=None)
        self.assertEqual(2, get_completed_builds(timedelta(hours=1)).count())

    def test_get_updated_builders(self):
        """
        get_updated_builders(timedelta(minutes=10)) should return all
        Lexbuilders which have been updated within the last 10 minutes.
        """
        builder1 = factory.make_lexbuilder()
        # This cheats the updated_at autonow value.
        Lexbuilder.objects.filter(pk=builder1.pk).update(
            updated_at=datetime.utcnow() - timedelta(minutes=12))
        builder2 = factory.make_lexbuilder()
        builder3 = factory.make_lexbuilder()
        self.assertEqual(2, get_updated_builders(timedelta(minutes=10)).count())
