# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from datetime import datetime, timedelta
import urllib

from django.contrib.auth.models import Permission
from django.core.urlresolvers import reverse
from django.test import TestCase

from django_group_access.models import AccessGroup

from offspring.enums import ProjectBuildStates
from offspring.web.queuemanager.models import (
    LaunchpadProject, ProjectGroup, Lexbuilder)
from offspring.web.queuemanager.forms import (
    LaunchpadProjectForm,
    ProjectGroupForm,
    CreateProjectForm
)
from offspring.web.queuemanager.tests.helpers import (
    create_build_results_with_durations)
from offspring.web.queuemanager.tests.factory import factory


class ProjectGroupsDetailsViewTests(TestCase):

    def setUp(self):
        self.project_group = factory.make_project_group()
        self.project1 = factory.make_project(project_group=self.project_group)
        self.project2 = factory.make_project(project_group=self.project_group)
        self.user = factory.make_user()
        self.view_url = reverse(
            "offspring.web.queuemanager.views.projectgroup_details",
            args=[self.project_group.name])

    def test_projectgroup_details_view_restricted(self):
        """
        Test project details view renders the restriction banner if at least
        one access groups exist.
        """
        access_group = AccessGroup.objects.create(name='Group')
        access_group.members.add(self.user)
        self.project2.access_groups.add(access_group)
        self.client.login(
            username=self.user.username, password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

    def test_projectgroup_details_view_not_restricted(self):
        """
        Test project details view does not render the restriction banner
        if no access group exists.
        """
        response = self.client.get(self.view_url)
        self.assertNotContains(
            response,
            'global-notification restricted-access-notification',
        )


class ProjectDetailsViewTests(TestCase):

    def setUp(self):
        self.project = factory.make_project()
        self.view_url = reverse(
            'offspring.web.queuemanager.views.project_details',
            args=[self.project.name])
        self.user = factory.make_user()

    def test_project_details_view(self):
        """
        Test project details view.
        """
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'Project Details for %s' % self.project.title.capitalize(),
            status_code=200,
            msg_prefix=response.content
        )

    def test_project_details_view_restricted(self):
        """
        Test project details view renders the restriction banner if at least
        one access group exists.
        """
        access_group = AccessGroup.objects.create(name='Group')
        access_group.members.add(self.user)
        self.project.access_groups.add(access_group)
        self.client.login(
            username=self.user.username, password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

    def test_project_details_view_not_restricted(self):
        """
        Test project details view does not render the restriction banner if
        no access groups exist.
        """
        response = self.client.get(self.view_url)
        self.assertNotContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

class AccessTests(TestCase):
    """
    Tests for record visibility.
    """
    def setUp(self):
        self.project1 = factory.make_project()
        self.project2 = factory.make_project()
        self.group, _ = AccessGroup.objects.get_or_create(name='Group')
        self.project2.access_groups.add(self.group)
        self.projects_url = reverse(
            'offspring.web.queuemanager.views.projects')
        self.builds_url = reverse(
            'offspring.web.queuemanager.views.builds')

    def test_project_visibility_anonymous_user(self):
        """
        Only unrestricted projects (with no access groups
        specified) are visibile to anonymous users.
        """
        response = self.client.get(self.projects_url)
        self.assertContains(response, self.project1.title)
        self.assertNotContains(response, self.project2.title)

    def test_project_visibility_authenticated(self):
        """
        Restricted projects are visibile to users in the correct
        group.
        """
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.projects_url)
        self.assertContains(response, self.project1.title)
        self.assertContains(response, self.project2.title)

    def test_build_result_visibility_anonymous_user(self):
        """
        Only builds for unrestricted projects are visible to
        anonymous users.
        """
        result = factory.make_build_result(self.project1)
        hidden_result = factory.make_build_result(self.project2)
        response = self.client.get(self.builds_url)
        self.assertContains(response, result.name)
        self.assertNotContains(response, hidden_result.name)

    def test_build_result_visibility_authenticated_user(self):
        """
        Builds for restricted projects are visibile to users in the correct
        group.
        """
        result = factory.make_build_result(self.project1)
        hidden_result = factory.make_build_result(self.project2)
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.builds_url)
        self.assertContains(response, result.name)
        self.assertContains(response, hidden_result.name)

    def test_build_request_visibility_anonymous_user(self):
        """
        Only build requests for unrestricted projects are visible to
        anonymous users.
        """
        request = factory.make_build_request(self.project1)
        hidden_request = factory.make_build_request(self.project2)
        response = self.client.get(self.projects_url)
        self.assertContains(response, str(request))
        self.assertNotContains(response, str(hidden_request))

    def test_build_request_visibility_authenticated_user(self):
        """
        Build requests for restricted projects are visibile to users in the
        correct group.
        """
        request = factory.make_build_request(self.project1)
        hidden_request = factory.make_build_request(self.project2)
        user = factory.make_user()
        self.group.members.add(user)
        self.client.login(username=user.username, password=user.username)
        response = self.client.get(self.projects_url)
        self.assertContains(response, str(request))
        self.assertContains(response, str(hidden_request))


class ProjectEditViewTests(TestCase):

    def setUp(self):
        self.project = factory.make_project()
        self.view_url = reverse(
            'offspring.web.queuemanager.views.project_edit',
            args=[self.project.name])
        self.user = factory.make_user()
        p = Permission.objects.get(codename="change_project")
        self.user.user_permissions.add(p)

    def test_project_edit_view(self):
        """
        Test project edit view with authenticated user.
        """
        self.client.login(
            username=self.user.username, password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'Update details for %s' % self.project.title.capitalize(),
            msg_prefix=response.content
        )


class BuildResultViewTests(TestCase):
    def setUp(self):
        self.user = factory.make_user()
        self.project = factory.make_project()
        self.build = factory.make_build_result(project=self.project)
        self.view_url = reverse(
            'offspring.web.queuemanager.views.build_details',
            args=[self.project.name, self.build.name])

    def test_buildresult_details_view(self):
        """
        Test BuildResult details view.
        """
        response = self.client.get(self.view_url)
        self.assertContains(
            response, 'Build Details', status_code=200,
            msg_prefix=response.content)

    def test_buildresult_details_view_restricted(self):
        """
        Test build result details view renders the restriction banner if at
        least one access group exists.
        """
        access_group = AccessGroup.objects.create(name='Group')
        access_group.members.add(self.user)
        self.project.access_groups.add(access_group)
        self.client.login(
            username=self.user.username, password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

    def test_buildresult_details_view_not_restricted(self):
        """
        Test build result details view does not render the restriction banner
        if no access groups exist.
        """
        response = self.client.get(self.view_url)
        self.assertNotContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

class BuildersListViewTests(TestCase):
    def setUp(self):
        user = factory.make_user()
        self.builder1 = factory.make_lexbuilder(
            name="BuilderA", is_retired=True)
        self.builder2 = factory.make_lexbuilder(name="BuilderB")
        self.client.login(
            username=user.username, password=user.username)

    def test_builderslist_does_not_show_retired_builders(self):
        """
        Test builders list does not show retired builders.
        """
        response = self.client.get(reverse("builders_list"))
        self.assertNotContains(
            response, self.builder1.name, status_code=200,
            msg_prefix=response.content)
        self.assertContains(
            response, self.builder2.name, status_code=200,
            msg_prefix=response.content)

class BuilderDetailsViewTests(TestCase):
    def setUp(self):
        user = factory.make_user()
        self.builder1 = factory.make_lexbuilder(
            name="BuilderA", is_retired=True)
        self.builder2 = factory.make_lexbuilder(name="BuilderB")
        self.client.login(
            username=user.username, password=user.username)

    def test_builderdetails_404s_for_retired_builders(self):
        """
        Test builder details are not available for retired builders.
        """
        response = self.client.get(reverse("builder_details", kwargs={'slug': self.builder1.name}))
        self.assertEqual(response.status_code, 404)
        response = self.client.get(reverse("builder_details", kwargs={'slug': self.builder2.name}))
        self.assertEqual(response.status_code, 200)

class BuildResultPublishViewTests(TestCase):

    def test_buildresult_publish_view(self):
        """
        Test build result publish view with authorized user.
        """
        project = factory.make_project()
        build = factory.make_build_result(
            project=project, result=ProjectBuildStates.SUCCESS)
        user = factory.make_user()
        p = Permission.objects.get(codename="add_release")
        user.user_permissions.add(p)
        self.assertTrue(
            self.client.login(username=user.username, password=user.username))
        response = self.client.get(
            reverse('offspring.web.queuemanager.views.build_publish',
                    args=[project.name, build.name]))
        self.assertContains(
            response, 'Release %s build' % project.title.capitalize(),
            status_code=200, msg_prefix=response.content)


class ProjectBuildsViewTests(TestCase):

    def setUp(self):
        self.user = factory.make_user()
        self.project = factory.make_project()
        self.build = factory.make_build_result(project=self.project)
        self.view_url = reverse(
            'offspring.web.queuemanager.views.builds',
            args=[self.project.name])

    def test_project_builds_view(self):
        """
        Test project build result list view.
        """
        response = self.client.get(self.view_url)
        self.assertContains(
            response, 'Build Records for %s' % self.project.title,
            status_code=200, msg_prefix=response.content)

    def test_project_builds_view_restricted(self):
        """
        Test project builds view renders the restriction banner if at least
        one access group exists.
        """
        access_group = AccessGroup.objects.create(name='Group')
        access_group.members.add(self.user)
        self.project.access_groups.add(access_group)
        self.client.login(
            username=self.user.username, password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )

    def test_project_builds_view_not_restricted(self):
        """
        Test project builds view does not render the restriction banner if
        no access groups exist.
        """
        response = self.client.get(self.view_url)
        self.assertNotContains(
            response,
            'global-notification restricted-access-notification',
            status_code=200,
            msg_prefix=response.content
        )


class OverallBuildsViewTests(TestCase):

    def setUp(self):
        self.user = factory.make_user()
        self.project = factory.make_project()
        self.build = factory.make_build_result(project=self.project)
        self.view_url = reverse('offspring.web.queuemanager.views.builds')

    def test_builds_view(self):
        """
        Test build result list view.
        """
        response = self.client.get(self.view_url)
        self.assertContains(
            response, self.build.name, status_code=200,
            msg_prefix=response.content)


class LaunchpadProjectViewsTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.launchpad_project_create")
        self.user = factory.make_user()

    def test_launchpad_project_create_view_redirects_unauthorized_user(self):
        """
        If the logged in user doesn't have permission to access the form, they
        should be directed to login.
        """
        self.client.login(username=self.user.username,
                          password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertRedirects(
            response,
            "/admin/?next=" + urllib.quote(self.view_url),
        )

    def test_get_launchpad_project_create_view(self):
        """
        If the logged in user has the correct permissions, a GET request
        should get the LaunchpadProjectForm.
        """
        p = Permission.objects.get(codename="add_launchpadproject")
        self.user.user_permissions.add(p)
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.view_url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], LaunchpadProjectForm),
            "Form is not LaunchpadProjectForm")
        self.assertTemplateUsed(
            response, "queuemanager/launchpad_project_create_popup.html")

    def test_post_launchpad_project_create_view(self):
        """
        A POST request should create a new LaunchpadProject with the entered
        name.
        """
        p = Permission.objects.get(codename="add_launchpadproject")
        self.user.user_permissions.add(p)
        self.client.login(username=self.user.username,
                          password=self.user.username)
        self.assertQuerysetEqual(LaunchpadProject.objects.all(), [])
        response = self.client.post(
            self.view_url, {"name": u"test"})
        self.assertEqual(response.status_code, 200)
        self.assertQuerysetEqual(
            LaunchpadProject.objects.all(), ['<LaunchpadProject: test>'])


class ProjectGroupFormTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.projectgroup_create")
        self.user = factory.make_user()

    def test_projectgroup_create_view_redirects_unauthorized_user(self):
        """
        If the logged in user doesn't have permission to access the form, they
        should be directed to login.
        """
        self.client.login(username=self.user.username,
                          password=self.user.username)
        response = self.client.get(self.view_url)
        self.assertRedirects(
            response,
            "/admin/?next=" + urllib.quote(self.view_url),
        )

    def test_get_projectgroup_create_view(self):
        """
        If the logged in user has the correct permissions, a GET request
        should get the ProjectGroupForm.
        """
        add_projectgroup_permission = Permission.objects.get(
            codename="add_projectgroup")
        self.user.user_permissions.add(add_projectgroup_permission)
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.view_url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], ProjectGroupForm),
            "Form is not ProjectGroupForm")
        self.assertTemplateUsed(
            response, "queuemanager/projectgroup_create_popup.html")

    def test_post_projectgroup_create_view(self):
        """
        A POST request should create a new ProjectGroup with the entered
        name and title.
        """
        add_projectgroup_permission = Permission.objects.get(
            codename="add_projectgroup")
        self.user.user_permissions.add(add_projectgroup_permission)
        self.client.login(username=self.user.username,
                          password=self.user.username)
        post_data = {"name": u"test", "title": u"Test ProjectGroup"}
        response = self.client.post(self.view_url, post_data)
        self.assertContains(response, text='opener.dismissAddAnotherPopup')
        self.assertEqual(ProjectGroup.objects.values()[0], post_data)


class CreateProjectTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.project_create")
        self.user = factory.make_user()

    def test_get_project_form_filters_projects_groups_for_user(self):
        """
        The ProjectForm should filter the list of ProjectGroups that a user can
        select from to ones that they already have projects in.
        """
        group1 = AccessGroup.objects.create(name="Group1")
        group2 = AccessGroup.objects.create(name="Group2")
        group1.members.add(self.user)

        project_group1 = factory.make_project_group()
        project_group2 = factory.make_project_group()

        project1 = factory.make_project(project_group=project_group1)
        project2 = factory.make_project(project_group=project_group1)
        project3 = factory.make_project(project_group=project_group2)

        for project, group in [(project1, group1), (project2, group2),
                               (project3, group2)]:
            project.access_groups.add(group)

        add_projectgroup_permission = Permission.objects.get(
            codename="add_project")
        self.user.user_permissions.add(add_projectgroup_permission)
        self.client.login(username=self.user.username,
                          password=self.user.username)

        response = self.client.get(self.view_url)
        self.assertEqual(200, response.status_code)
        self.assertTrue(
            isinstance(response.context["form"], CreateProjectForm),
            "Form is not a CreateProjectForm")
        self.assertTemplateUsed(
            response, "queuemanager/project_create.html")
        self.assertContains(response, '<option value="%s">' %
                            project_group1.name)
        self.assertNotContains(response, '<option value="%s">' %
                            project_group2.name)


class AliveBuildsViewTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.alive_builds")
        self.user = factory.make_user()
        project = factory.make_project()
        started_at = datetime.utcnow() - timedelta(hours=2)
        factory.make_build_result(
            project=project, requestor=None,
            result=ProjectBuildStates.SUCCESS,
            finished_at=started_at + timedelta(minutes=30),
            builder=None)
        create_build_results_with_durations(
            project, [50, 50], result=ProjectBuildStates.SUCCESS)

    def test_alive_returns_build_count(self):
        """
        Fetching the /alive/builds/ URL returns the number of builds completed
        in the last hour.
        """
        response = self.client.get(self.view_url)
        self.assertEqual("2", response.content)

    def test_alive_with_optional_duration(self):
        """
        It's possible to specify a duration, which is used as a timedelta for
        the alive url.
        """
        response = self.client.get(self.view_url, {"hours": 2})
        self.assertEqual("3", response.content)


class AliveBuildersViewTests(TestCase):

    def setUp(self):
        self.view_url = reverse(
            "offspring.web.queuemanager.views.alive_builders")
        builder1 = factory.make_lexbuilder()
        Lexbuilder.objects.filter(pk=builder1.pk).update(
            updated_at=datetime.utcnow() - timedelta(minutes=12))
        builder2 = factory.make_lexbuilder()
        builder3 = factory.make_lexbuilder()
        builder4 = factory.make_lexbuilder(machine_type="armel")
        builder5 = factory.make_lexbuilder(machine_type="armel")
        Lexbuilder.objects.filter(pk=builder5.pk).update(
            updated_at=datetime.utcnow() - timedelta(minutes=12))

    def test_alive_returns_updates_builders_count(self):
        """
        Fetching the /alive/builders/ URL returns the number of builders
        updated within the 10 minutes.
        """
        response = self.client.get(self.view_url)
        self.assertEqual("3", response.content)

    def test_alive_with_optional_duration(self):
        """
        It's possible to specify a duration, which is used as a timedelta for
        the alive url.
        """
        response = self.client.get(self.view_url, {"minutes": 15})
        self.assertEqual("5", response.content)

    def test_alive_with_optional_arch_x86_64(self):
        """
        It's possible to specify an architecture, which is used to determine
        what the state of a whole class of builders is.
        """
        response = self.client.get(self.view_url, {"arch": "x86_64"})
        self.assertEqual("2", response.content)

    def test_alive_with_optional_arch_armel(self):
        """
        It's possible to specify an architecture, which is used to determine
        what the state of a whole class of builders is.
        """
        response = self.client.get(self.view_url, {"arch": "armel"})
        self.assertEqual("1", response.content)

    def test_alive_with_optional_duration_and_optional_arch_armel(self):
        """
        It's possible to specify a duration and an architecture together.
        """
        response = self.client.get(
            self.view_url, {"minutes": 15, "arch": "armel"})
        self.assertEqual("2", response.content)
